package com.sun.xml.internal.ws.api;

import java.security.AccessController;
import java.lang.reflect.Field;
import java.security.PrivilegedAction;
import com.sun.xml.internal.ws.client.WSServiceDelegate;
import java.net.URL;
import com.sun.istack.internal.Nullable;
import java.util.Iterator;
import com.sun.istack.internal.NotNull;
import com.sun.xml.internal.ws.api.server.Container;
import javax.xml.bind.JAXBContext;
import javax.xml.ws.Dispatch;
import javax.xml.ws.Service;
import javax.xml.namespace.QName;
import javax.xml.ws.WebServiceFeature;
import com.sun.xml.internal.ws.api.addressing.WSEndpointReference;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.Set;
import javax.xml.ws.spi.ServiceDelegate;

public abstract class WSService extends ServiceDelegate implements ComponentRegistry
{
    private final Set<Component> components;
    protected static final ThreadLocal<WSService.InitParams> INIT_PARAMS;
    protected static final WSService.InitParams EMPTY_PARAMS;
    
    protected WSService() {
        this.components = new CopyOnWriteArraySet();
    }
    
    public abstract <T> T getPort(WSEndpointReference p0, Class<T> p1, WebServiceFeature... p2);
    
    public abstract <T> Dispatch<T> createDispatch(QName p0, WSEndpointReference p1, Class<T> p2, Service.Mode p3, WebServiceFeature... p4);
    
    public abstract Dispatch<Object> createDispatch(QName p0, WSEndpointReference p1, JAXBContext p2, Service.Mode p3, WebServiceFeature... p4);
    
    @NotNull
    public abstract Container getContainer();
    
    @Nullable
    public <S> S getSPI(@NotNull Class<S> spiType) {
        for (Component c : this.components) {
            S s = c.getSPI(spiType);
            if (s != null) {
                return s;
            }
            continue;
        }
        return this.getContainer().getSPI(spiType);
    }
    
    @NotNull
    public Set<Component> getComponents() {
        return this.components;
    }
    
    public static WSService create(URL wsdlDocumentLocation, QName serviceName) {
        return new WSServiceDelegate(wsdlDocumentLocation, serviceName, Service.class, new WebServiceFeature[0]);
    }
    
    public static WSService create(QName serviceName) {
        return create(null, serviceName);
    }
    
    public static WSService create() {
        return create(null, new QName(WSService.class.getName(), "dummy"));
    }
    
    public static Service create(URL wsdlDocumentLocation, QName serviceName, WSService.InitParams properties) {
        if (WSService.INIT_PARAMS.get() != null) {
            throw new IllegalStateException("someone left non-null InitParams");
        }
        WSService.INIT_PARAMS.set(properties);
        try {
            Service svc = Service.create(wsdlDocumentLocation, serviceName);
            if (WSService.INIT_PARAMS.get() != null) {
                throw new IllegalStateException("Service " + svc + " didn't recognize InitParams");
            }
            return svc;
        }
        finally {
            WSService.INIT_PARAMS.set(null);
        }
    }
    
    public static WSService unwrap(final Service svc) {
        return (WSService)AccessController.doPrivileged(new PrivilegedAction<WSService>() {
            @Override
            public WSService run() {
                try {
                    Field f = svc.getClass().getField("delegate");
                    f.setAccessible(true);
                    Object delegate = f.get(svc);
                    if (!(delegate instanceof WSService)) {
                        throw new IllegalArgumentException();
                    }
                    return (WSService)delegate;
                }
                catch (NoSuchFieldException e) {
                    AssertionError x = new AssertionError("Unexpected service API implementation");
                    x.initCause(e);
                    throw x;
                }
                catch (IllegalAccessException e2) {
                    IllegalAccessError x2 = new IllegalAccessError(e2.getMessage());
                    x2.initCause(e2);
                    throw x2;
                }
            }
        });
    }
    
    static {
        INIT_PARAMS = new ThreadLocal();
        EMPTY_PARAMS = new WSService.InitParams();
    }
    
    public static final class InitParams
    {
        private Container container;
        
        public void setContainer(Container c) {
            this.container = c;
        }
        
        public Container getContainer() {
            return this.container;
        }
    }
}
